#include "common.h"
#include "const.h"
#include <iostream>
#include <fstream>
#include <cmath>
#include <cstring>
#include <cstdlib>

#include <Eigen/Dense>

using namespace Eigen;
using namespace std;

void Averaging_Sinr(int ue_idx, double sinr);
double Get_distance(LOCATION, LOCATION);

/*===================================================================
FUNCTION: UE::Receive_DL( void )

AUTHOR: SJ, KR, DW

DESCRIPTION: UE check the received resource block and check if it has failed or not.

NOTES:


===================================================================*/
void MS::Receive_DL(int _ms_idx)
{

	receive_self_ms_idx = _ms_idx;

	// initialize capacity
	_capacity = 0;


	Find_Allocated_Rbs_Mcs(receive_self_ms_idx);
	
	/*
	cout << "self_ms_idx = " << self_ms_idx << endl;
	cout << "rb_indices.size() = " << rb_indices.size() << endl;
	getchar();
	*/
	
	if (rb_indices.size() > 0)
	{
		Rate_matching();
		Compute_Num_Bin_Symbol();
		Compute_Code_Rate();
		//Compute_lls_mapping();


		////////// Vienna BLER mapping //////////////////
		/////////////////////////////////////////////////

		int mapping_snr_dB_idx = 0;
		int mapping_cqi_idx = 0;

		double BLER_value = 0.;

		//double x_axis = linear2dB(_input2table.snr);

		double x_axis = received_sinr_avg;

		//double x_axis = 10*log10(ESINR_linear);


		for (int snr_idx = 0; snr_idx < 612; snr_idx++)
		{

			if (SNR_dB[snr_idx] > x_axis)
			{
				//cout << "snr_idx = " << snr_idx << endl;
				mapping_snr_dB_idx = snr_idx - 1;
				break;
			}

			else  
			{
				mapping_snr_dB_idx = 611;
			}
		}
		/////////////////////////////////////////////////////////////////////////
		/////////////////////////////////////////////////////////////////////////
		//cout << "mapping_snr_dB_idx = " << mapping_snr_dB_idx << endl;
		//cout << "_cqi_idx = " << _cqi_idx << endl;

		BLER_value = (BLER_vienna[mapping_snr_dB_idx][_cqi_idx - 1]);

		//cout << "BLER_value = " << BLER_value << endl;
		//getchar();

		BLER_SUM_Value = BLER_SUM_Value + BLER_value;
		BLER_SUM_count = BLER_SUM_count + 1;


		if (t > N_pf)
		{
			AddThroughput(BLER_value);

			// BLER curve LTE ̹Ƿ ϴ ׳ 0.1 
			//AddThroughput(0.1);
		}
		
			
	}
	
	

}

/*===================================================================
FUNCTION: UE::AddThroughput( void )

AUTHOR: SJ, KR, DW

DESCRIPTION:

NOTES:

===================================================================*/
void MS::AddThroughput(double prob)
{
	//for check
	if (prob < 0 || prob > 1)
	{
		cout << "something wrong with prob_failure at ue.cpp. prob = " << prob; getchar();
	}

	double tossing_coin = randnum.u();	



	if (link[receive_self_ms_idx].num_re_tx == 0)
	{
		total_num_tx++;
	}




	if (prob < tossing_coin || isnan(prob)) // Tossing a biased coin
	{
		link[receive_self_ms_idx].ACK = true;

		link[receive_self_ms_idx]._throughput += _info_bits;


		link[receive_self_ms_idx].num_re_tx = 0;

	}
	else
	{
		link[receive_self_ms_idx].ACK = false;


		if (link[receive_self_ms_idx].num_re_tx == 0)
		{
			failed++;
		}


		link[receive_self_ms_idx].num_re_tx++;



		if (link[receive_self_ms_idx].num_re_tx > 4)
		{
			link[receive_self_ms_idx].num_re_tx = 0;
			link[receive_self_ms_idx].ACK = true;
		}


	}
}

double MS::Return_initial_packet_error(void)
{
	//return ((double)failed / (double)total_num_tx);

	if (total_num_tx == 0)  
	{
		return 0;
	}
	else
	{
		return ((double)failed / (double)total_num_tx);
	}


}

/*===================================================================
FUNCTION: UE::Find_allocated_rbs( void )

AUTHOR: DH

DESCRIPTION: UE checks the schedule map, and if allocated rbs is found,
add rbs index to vector ('rb_indices') and check the mcs type.

NOTES: Only used by UE::receiveUL( void )

INPUT: schedule_map,
OUTPUT: rb_indices


===================================================================*/
void MS::Find_Allocated_Rbs_Mcs(int self_ms_idx)
{

	int temp_ms_idx = self_ms_idx; //ms_idx (0~num_MS) from main function

	SCHEDULE_DECISION decision_of_interest;

	rb_indices.clear(); //vector <int> rb_indices

	int temp_mcs = 0;
	int temp_cqi = 0;

	for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
	{
		for (int stream = 0; stream < mx_ue_mumimo; stream++)
		{
			decision_of_interest = pppSchedule_Map[link[temp_ms_idx]._sector_in_control][freq_idx][stream];
			
			if (decision_of_interest.ue_selected == temp_ms_idx) 
			{
				rb_indices.push_back(freq_idx);

				if (rb_indices.size() == 1)
				{
					if (temp_mcs == -1)
					{
						cout << "error!!" << endl;
						getchar();
					}

					temp_mcs = decision_of_interest.mcs_selected;
					temp_cqi = decision_of_interest.cqi_selected;
					_capacity = decision_of_interest.capacity;


				}
				else
				{
					if (temp_mcs != decision_of_interest.mcs_selected)
					{
						cerr << "Something Wrong in the Scheduler" << endl;
						getchar();
					}

					if (temp_cqi != decision_of_interest.cqi_selected)
					{
						cerr << "Something Wrong in the Scheduler CQI" << endl;
						getchar();
					}
				}
				
			}
			
		}
	}

	
	//if ACK, temp_mcs from schedule map -> _mcs_idx 
	if (link[temp_ms_idx].ACK)
	{

		if (_mcs_idx == -1)
		{
			cout << "error!!" << endl;
			getchar();
		}

		_mcs_idx = temp_mcs;
		_cqi_idx = temp_cqi;
	}


	

}


/*===================================================================
FUNCTION: UE::Compute_num_bin_symbol( void )

NOTES: 2009-12-30

===================================================================*/
void MS::Rate_matching(void)
{
	if (link[receive_self_ms_idx].ACK)
	{
		if (true)
		{
			int _num_traffic_ = 0;
			int _info_bits_ = 0;

			int _mod_type_;

			if (_mcs_idx <= 9)       { _mod_type_ = QPSK; }
			else if (_mcs_idx <= 16) { _mod_type_ = QAM16; }
			else if (_mcs_idx <= 28) { _mod_type_ = QAM64; }
			else                      { cerr << "Something Wrong - rate matching" << endl; }


			int N_DMRS = 0;
			int N_overhead = 0;

			int N_RE_a;
			int N_RE;
			int N_info;

			N_RE_a = num_freq_per_rbs * num_ofdm_symbols_per_subband_per_1ms - N_DMRS - N_overhead;

			N_RE = MIN(156, N_RE_a) * rb_indices.size();


			double coding_rate_x1024 = 0;
			double coding_rate;
			int TB_Size_bits = 0;

			///////
			if (_mcs_idx == 0)        { coding_rate_x1024 = 120; }
			else if (_mcs_idx == 1)   { coding_rate_x1024 = 157; }
			else if (_mcs_idx == 2)   { coding_rate_x1024 = 193; }
			else if (_mcs_idx == 3)   { coding_rate_x1024 = 251; }
			else if (_mcs_idx == 4)   { coding_rate_x1024 = 308; }
			else if (_mcs_idx == 5)   { coding_rate_x1024 = 379; }
			else if (_mcs_idx == 6)   { coding_rate_x1024 = 449; }
			else if (_mcs_idx == 7)   { coding_rate_x1024 = 526; }
			else if (_mcs_idx == 8)   { coding_rate_x1024 = 602; }
			else if (_mcs_idx == 9)   { coding_rate_x1024 = 679; }
			else if (_mcs_idx == 10)  { coding_rate_x1024 = 340; }
			else if (_mcs_idx == 11)  { coding_rate_x1024 = 378; }
			else if (_mcs_idx == 12)  { coding_rate_x1024 = 434; }
			else if (_mcs_idx == 13)  { coding_rate_x1024 = 490; }
			else if (_mcs_idx == 14)  { coding_rate_x1024 = 553; }
			else if (_mcs_idx == 15)  { coding_rate_x1024 = 616; }
			else if (_mcs_idx == 16)   { coding_rate_x1024 = 658; }
			else if (_mcs_idx == 17)   { coding_rate_x1024 = 438; }
			else if (_mcs_idx == 18)   { coding_rate_x1024 = 466; }
			else if (_mcs_idx == 19)   { coding_rate_x1024 = 517; }
			else if (_mcs_idx == 20)   { coding_rate_x1024 = 567; }
			else if (_mcs_idx == 21)   { coding_rate_x1024 = 616; }
			else if (_mcs_idx == 22)   { coding_rate_x1024 = 719; }
			else if (_mcs_idx == 23)   { coding_rate_x1024 = 772; }
			else if (_mcs_idx == 24)   { coding_rate_x1024 = 822; }
			else if (_mcs_idx == 25)  { coding_rate_x1024 = 873; }
			else if (_mcs_idx == 26)  { coding_rate_x1024 = 910; }
			else if (_mcs_idx == 27)  { coding_rate_x1024 = 948; }
			else
			{
				cout << "Something Wrong rate matching" << endl;
				getchar();
			}

			coding_rate = coding_rate_x1024 / 1024.;

			/// overhead Ȯؼ ڵ  ó ϴ غ
			N_info = (int)(N_RE * coding_rate * _mod_type * NUM_UE_Layer* (1. - overhead));

			if (N_info <= 3834)
			{
				int n = MAX(3, floor(log2(N_info)) - 6.);

				int N_info_a = MAX(24, pow(2, n) * floor(N_info / pow(2, n)));

				_info_bits_ = N_info_a;
			}
			else
			{
				int n = floor(log2(N_info - 24)) - 5;

				int N_info_a = MAX(3840, pow(2, n) * round(((N_info - 24) / pow(2, n))));

				if (coding_rate <= (1. / 4.))
				{
					int C = ceil((N_info_a + 24) / 3816);

					_info_bits_ = 8 * C * ceil((N_info_a + 24) / (8 * C)) - 24;
				}
				else
				{
					if (N_info_a >= 8424)
					{
						int C = ceil((N_info_a + 24) / 8424);

						_info_bits_ = 8 * C * ceil((N_info_a + 24) / (8 * C)) - 24;
					}
					else
					{
						_info_bits_ = 8 * ceil((N_info_a + 24) / 8) - 24;
					}
				}

			}


			if (((double)_info_bits_ / (double)_num_traffic_) > 0.93)
			{
				//_mcs_idx = _mcs_idx - 1;
			}
		}

		else if (false)  //// old code - use MCS index 
		{
			int _num_traffic_ = 0;
			int _info_bits_ = 0;

			int _mod_type_; 

			if (_mcs_idx <= 9)       { _mod_type_ = QPSK;  }
			else if (_mcs_idx <= 16) { _mod_type_ = QAM16; }
			else if (_mcs_idx <= 28) { _mod_type_ = QAM64; }
			else                      { cerr << "Something Wrong" << endl; }


			_num_traffic_ = (int)(num_freq_per_rbs * num_ofdm_symbols_per_subband_per_1ms * rb_indices.size() * (1. - overhead) * _mod_type_); //_mod_type_ -> _mod_type


			int _i_tbs;

			_i_tbs = _mcs_idx;




			if (_i_tbs < 10)      { _i_tbs = _i_tbs; }
			else if (_i_tbs < 17) { _i_tbs -= 1; }
			else if (_i_tbs < 29) { _i_tbs -= 2; }
			else                  {
				cout << "Something Wrong tbs index" << endl;
				getchar();
			}



		
			_info_bits_ = TBS[rb_indices.size()][_i_tbs] + num_crc;



			if (((double)_info_bits_ / (double)_num_traffic_) > 0.93) 
			{
				_mcs_idx = _mcs_idx - 1;
			}

		}
		else  
		{
			int _num_traffic_;
			int _info_bits_;

			int _mod_type_; 

			///// USE CQI index
			if (_mcs_idx <= 6) { _mod_type_ = QPSK; }
			else if (_mcs_idx <= 9) { _mod_type_ = QAM16; }
			else if (_mcs_idx <= 15) { _mod_type_ = QAM64; }
			else                      { cerr << "Something Wrong" << endl; }


			_num_traffic_ = (int)(num_freq_per_rbs * num_ofdm_symbols_per_subband_per_1ms * rb_indices.size() * (1. - overhead) * _mod_type_); //_mod_type_ -> _mod_type

			//_num_traffic_ = (int)(num_freq_per_rbs * num_ofdm_symbols_per_subband_per_1ms * rb_indices.size() * _mod_type_); //_mod_type_ -> _mod_type


			double coding_rate_x1024 = 0;
			double coding_rate;
			int TB_Size_bits = 0;

			/////// Use TR 36.213 Table 7.2.3-1 : CQI Table (Release 8)
			if (_mcs_idx == 1)   { coding_rate_x1024 = 78; }
			else if (_mcs_idx == 2)   { coding_rate_x1024 = 120; }
			else if (_mcs_idx == 3)   { coding_rate_x1024 = 193; }
			else if (_mcs_idx == 4)   { coding_rate_x1024 = 308; }
			else if (_mcs_idx == 5)   { coding_rate_x1024 = 449; }
			else if (_mcs_idx == 6)   { coding_rate_x1024 = 602; }
			else if (_mcs_idx == 7)   { coding_rate_x1024 = 378; }
			else if (_mcs_idx == 8)   { coding_rate_x1024 = 490; }
			else if (_mcs_idx == 9)   { coding_rate_x1024 = 616; }
			else if (_mcs_idx == 10)  { coding_rate_x1024 = 466; }
			else if (_mcs_idx == 11)  { coding_rate_x1024 = 567; }
			else if (_mcs_idx == 12)  { coding_rate_x1024 = 666; }
			else if (_mcs_idx == 13)  { coding_rate_x1024 = 772; }
			else if (_mcs_idx == 14)  { coding_rate_x1024 = 873; }
			else if (_mcs_idx == 15)  { coding_rate_x1024 = 948; }
			else
			{
				cout << "Something Wrong CQI index decision" << endl;
				getchar();
			}



			int sym_per_RB = num_freq_per_rbs * (num_ofdm_symbols_per_subband_per_1ms / 2) - 4;  // sym_per_RB = 80 , 4 = RS symbol???

			coding_rate = coding_rate_x1024 / 1024;


			TB_Size_bits = (int)(8 * round((sym_per_RB * rb_indices.size() * _mod_type_ * coding_rate * 2) / 8.) - 24);





			_info_bits_ = TB_Size_bits + num_crc;




		}

		
		
	}

}

/*===================================================================
FUNCTION: UE::Compute_num_bin_symbol( void )

AUTHOR: SJ, KR, DW

DESCRIPTION:

NOTES:

===================================================================*/
void MS::Compute_Num_Bin_Symbol(void)
{
	// initial condition

	num_rx_rb = 0;

	if (link[receive_self_ms_idx].ACK == 0) // Previously failed case
	{
		for (int freq_idx = 0; freq_idx < (int)rb_indices.size(); freq_idx++)
		{
			double rx_sinr = 0;

			if (Scheduling_Type == 0) // roundrobin 1x2
			{
				rx_sinr = compute_tone_SINR_MRC(rb_indices[freq_idx]); //
				//rx_sinr = compute_tone_SINR(rb_indices[freq_idx]); //
			}
			else  
			{
				rx_sinr = compute_tone_SINR(rb_indices[freq_idx]); //

				//rx_sinr = compute_tone_SINR_MRC(rb_indices[freq_idx]); 

			}

			if (rx_sinr)
			{
				rbs_rx.push_back(rx_sinr);
			}
		}

		_num_traffic = (int)(num_freq_per_rbs * num_ofdm_symbols_per_subband_per_1ms * rbs_rx.size() * (1. - overhead) * _mod_type);

	}
	else // Previously successed or initial transmission case
	{

		rbs_rx.clear();

		for (int freq_idx = 0; freq_idx < (int)rb_indices.size(); freq_idx++)
		{
			double rx_sinr = 0;

			if (Scheduling_Type == 0)
			{
				rx_sinr = compute_tone_SINR_MRC(rb_indices[freq_idx]);
			}
			else
			{
				rx_sinr = compute_tone_SINR(rb_indices[freq_idx]);
			}

			if (rx_sinr)
			{
				rbs_rx.push_back(rx_sinr);
			}

		}

		if (_mcs_idx <= 9)       { _mod_type = QPSK; }
		else if (_mcs_idx <= 16) { _mod_type = QAM16; }
		else if (_mcs_idx <= 28) { _mod_type = QAM64; }
		else                      { cerr << "Something Wrong - compute bin sym" << endl; }



		int N_DMRS = 0;
		int N_overhead = 0;

		int N_RE_a;
		int N_RE;
		int N_info;

		N_RE_a = num_freq_per_rbs * num_ofdm_symbols_per_subband_per_1ms - N_DMRS - N_overhead;


		N_RE = MIN(156, N_RE_a) * rb_indices.size();


		double coding_rate_x1024 = 0;
		double coding_rate;
		int TB_Size_bits = 0;

		///////
		if (_mcs_idx == 0)        { coding_rate_x1024 = 120; }
		else if (_mcs_idx == 1)   { coding_rate_x1024 = 157; }
		else if (_mcs_idx == 2)   { coding_rate_x1024 = 193; }
		else if (_mcs_idx == 3)   { coding_rate_x1024 = 251; }
		else if (_mcs_idx == 4)   { coding_rate_x1024 = 308; }
		else if (_mcs_idx == 5)   { coding_rate_x1024 = 379; }
		else if (_mcs_idx == 6)   { coding_rate_x1024 = 449; }
		else if (_mcs_idx == 7)   { coding_rate_x1024 = 526; }
		else if (_mcs_idx == 8)   { coding_rate_x1024 = 602; }
		else if (_mcs_idx == 9)   { coding_rate_x1024 = 679; }
		else if (_mcs_idx == 10)  { coding_rate_x1024 = 340; }
		else if (_mcs_idx == 11)  { coding_rate_x1024 = 378; }
		else if (_mcs_idx == 12)  { coding_rate_x1024 = 434; }
		else if (_mcs_idx == 13)  { coding_rate_x1024 = 490; }
		else if (_mcs_idx == 14)  { coding_rate_x1024 = 553; }
		else if (_mcs_idx == 15)  { coding_rate_x1024 = 616; }
		else if (_mcs_idx == 16)   { coding_rate_x1024 = 658; }
		else if (_mcs_idx == 17)   { coding_rate_x1024 = 438; }
		else if (_mcs_idx == 18)   { coding_rate_x1024 = 466; }
		else if (_mcs_idx == 19)   { coding_rate_x1024 = 517; }
		else if (_mcs_idx == 20)   { coding_rate_x1024 = 567; }
		else if (_mcs_idx == 21)   { coding_rate_x1024 = 616; }
		else if (_mcs_idx == 22)   { coding_rate_x1024 = 719; }
		else if (_mcs_idx == 23)   { coding_rate_x1024 = 772; }
		else if (_mcs_idx == 24)   { coding_rate_x1024 = 822; }
		else if (_mcs_idx == 25)  { coding_rate_x1024 = 873; }
		else if (_mcs_idx == 26)  { coding_rate_x1024 = 910; }
		else if (_mcs_idx == 27)  { coding_rate_x1024 = 948; }
		else
		{
			cout << "Something Wrong compute bin sym" << endl;
			cout << "receive_self_ms_idx = " << receive_self_ms_idx << endl;
			cout << "_mcs_idx = " << _mcs_idx << endl;
			getchar();
		}

		coding_rate = coding_rate_x1024 / 1024.;



		/// overhead Ȯؼ ڵ  ó ϴ غ
		N_info = (int)(N_RE * coding_rate * _mod_type * NUM_UE_Layer* (1. - overhead));
		//N_info = (int)(N_RE * coding_rate * _mod_type * 1 * (1. - overhead));



		//N_info = (int)(N_RE * coding_rate * _mod_type * mx_ue_mumimo);

		if (N_info <= 3834)
		{
			int n = MAX(3, floor(log2(N_info)) - 6.);

			int N_info_a = MAX(24, pow(2, n) * floor(N_info / pow(2, n)));

			_info_bits = N_info_a;
		}
		else
		{
			int n = floor(log2(N_info - 24)) - 5;

			int N_info_a = MAX(3840, pow(2, n) * round(((N_info - 24) / pow(2, n))));

			if (coding_rate <= (1. / 4.))
			{
				int C = ceil((N_info_a + 24) / 3816);

				_info_bits = 8 * C * ceil((N_info_a + 24) / (8 * C)) - 24;
			}
			else
			{
				if (N_info_a >= 8424)
				{
					int C = ceil((N_info_a + 24) / 8424);

					_info_bits = 8 * C * ceil((N_info_a + 24) / (8 * C)) - 24;
				}
				else
				{
					_info_bits = 8 * ceil((N_info_a + 24) / 8) - 24;
				}
			}

		}


		_num_traffic = (int)(num_freq_per_rbs * num_ofdm_symbols_per_subband_per_1ms * rbs_rx.size() * (1. - overhead) * _mod_type);

	}


	double sinr = 0;
	for (int rb_idx = 0; rb_idx < (int)rbs_rx.size(); rb_idx++)
	{
		sinr += rbs_rx[rb_idx];
	}
	

	received_sinr_avg = 10. * log10(sinr / (double)num_rbs_in_1ms);     ///// dB scale

	/*
	cout << "receive_self_ms_idx = " << receive_self_ms_idx << endl;
	cout << "received_sinr_avg = " << received_sinr_avg << endl;
	getchar();
	*/




	//Averaging_Sinr(_self_idx, sinr / (double)num_rbs_in_1ms);

	Averaging_Sinr(receive_self_ms_idx, sinr / (double)num_rbs_in_1ms); 

}



/*===================================================================
FUNCTION: UE::Compute_Code_Rate( void )

AUTHOR: SJ, KR, DW

DESCRIPTION: Compute the code rate

NOTES:

===================================================================*/
void MS::Compute_Code_Rate(void)
{

	// compute ESINR
	// EESM method


	double e_sir = 0.;
	for (int i = 0; i < rbs_rx.size(); i++)
	{
		e_sir += exp(-rbs_rx[i] / Beta[_cqi_idx]);
	}

	e_sir = -Beta[_cqi_idx] * log(e_sir / (double)(rbs_rx.size()));
	
	ESINR_linear = e_sir;


	/*
	cout << "received_sinr_avg(dB) = " << received_sinr_avg << endl;
	cout << "ESINR_linear(dB) = " << 10*log10(ESINR_linear) << endl;
	getchar();
	*/


	
	/*
	_code_rate = ((double)_info_bits / (double)_num_traffic);
	_snr_scale_factor = 1;

	if (_code_rate < 0.3333333333333 && _mod_type != QPSK)
	{
		_code_rate = (double)1 / (double)3;
		_snr_scale_factor = ((double)_num_traffic / (double)(_info_bits * inverse_TC_rate));
	}

	if (_code_rate < 0.083333333 && _mod_type == QPSK)
	{
		_code_rate = (double)1 / (double)3;
		_snr_scale_factor = ((double)_num_traffic / (double)(_info_bits * inverse_TC_rate));
	}

	if (_code_rate < 1 && _code_rate >= 4. / 5. && _mod_type != QAM64) // 
	{
		_code_rate = 4. / 5.;
	}

	if (_code_rate < 1 && _code_rate >= 0.916666667 && _mod_type == QAM64) // 
	{
		_code_rate = 0.916666667;
	}



	if (_code_rate > 1)
	{
		
		int i_tbs = _mcs_idx;


		if (i_tbs < 10)      { i_tbs = i_tbs; }
		else if (i_tbs < 17) { i_tbs -= 1; }
		else if (i_tbs < 29) { i_tbs -= 2; }
		else                  {
			cout << "Something Wrong tbs index" << endl;
			getchar();
		}

		cout << " _num_traffic = " << _num_traffic << endl;
		cout << " rbs_rx.size() = " << rbs_rx.size() << endl;
		cout << " i_tbs        = " << i_tbs << endl;
		cout << " _info_bits   = " << _info_bits << endl;
		cout << " _code_rate   = " << _code_rate << endl;
		getchar();
	}


	_input2table.effective_code_rate = _code_rate;


	
	*/

	





}






/*===================================================================
FUNCTION: UE::Compute_lls_mapping( void )

AUTHOR:

DESCRIPTION:

NOTES:

===================================================================*/
void MS::Compute_lls_mapping()
{
	INPUT_TO_LLS_MAPPING set_mapping_info = Return_mapping_info();

	if (lls_mapping == 0)
	//if (true)
	{
		_input2table = eesm_methology(set_mapping_info, receive_self_ms_idx);
		//eesm_methology( set_mapping_info , _self_idx ) ;
	}
	else if (lls_mapping == 1) 
	{
		_input2table = rbir_methology(set_mapping_info, receive_self_ms_idx); 

		//rbir_methology( set_mapping_info , _self_idx ) ;
	}
	else if (lls_mapping == 2)
	{
		_input2table = mib_methology(set_mapping_info, receive_self_ms_idx);
	}
	else
	{
		cout << "There some problem at Compute_lls_mapping. Check!" << getchar();
	}
}

/*===================================================================
FUNCTION: UE::return_throughput( void )

AUTHOR: SJ

DESCRIPTION: return the total throughput over the frequency

NOTES:

===================================================================*/
INPUT_TO_LLS_MAPPING MS::Return_mapping_info(void)
{
	INPUT_TO_LLS_MAPPING input;
	input.mod_type = _mod_type;
	input.code_rate = _code_rate;
	input.snr_scale_factor = _snr_scale_factor;

	return (input);
}


/*===================================================================
FUNCTION: UE::compute_tone_SINR( )

DESCRIPTION:
===================================================================*/
double MS::compute_tone_SINR(int rb_idx)
{
	double Received_SINR = 0;
	double I_power = 0;
	MatrixXcd A, B;
	int stream_num = -1;

	for (int i = 0; i < mx_ue_mumimo; i++)
	{
		if (ppSchedulerRead[rb_idx][i].ue_selected == receive_self_ms_idx)
		{
			stream_num = i;
		}
	}
	if (stream_num > -1)
	{
		///cout << "receive_self_ms_idx = " << receive_self_ms_idx << endl;
		///cout << "stream_num = " << stream_num << endl;
		///getchar();


		num_rx_rb++;

		MatrixXcd w, u, h_u;

		MatrixXcd H_bar;



		double geometry = link[receive_self_ms_idx].str_signal - link[receive_self_ms_idx].interference; 

		double linear_signal = pow(10., link[receive_self_ms_idx].str_signal/10.);
		double linear_interference = pow(10., link[receive_self_ms_idx].interference/10.);



		H_bar = sqrt(linear_signal) * H_m[rb_idx] * sector[link[receive_self_ms_idx]._sector_in_control].W[rb_idx];
	/// Rx2                              RxT                    Tx2
		
		

		/*
		cout << "H_bar = " << H_bar << endl;
		cout << "H_m[rb_idx] = " << H_m[rb_idx] << endl;
		cout << "sector[link[receive_self_ms_idx]._sector_in_control].W[rb_idx] = " << sector[link[receive_self_ms_idx]._sector_in_control].W[rb_idx] << endl;
		*/




		h_u = (H_bar.col(stream_num));
//      Rx1

		// u = receiver matrix
		u = MMSE_Receiver(rb_idx, stream_num); // 1x2 //MESC_Receiver_Filter( rb_idx );
///    1xR

		//cout << "After MMSE" << endl;


		// u = receiver matrix
		A = u * h_u; // 
///        1xR  Rx1 


		///cout << "A = " << A << endl;
		///cout << "A.norm = " << A.norm() << endl;



		for (int stream = 0; stream < mx_ue_mumimo; stream++)
		{
			if (stream != stream_num && ppSchedulerRead[rb_idx][stream].ue_selected > -1)
			{
				MatrixXcd h_i = MatrixXcd::Zero(NUM_RX_Port,1);

				
				h_i = (H_bar.col(stream));
			//           Rx1

				B = u * h_i;
			//1x1  1xR Rx1



				////// MU mimo õ   interference 

				//I_power += B.norm() * B.norm(); // power_i *


			}
		}


		
		// u = receiver matrix


		I_power += linear_interference * (u * u.adjoint()).norm();


		///cout << "I_power 2222 = " << I_power << endl;

		Received_SINR = A.norm() * A.norm() / I_power; // power *

		/*
		cout << "A.norm() * A.norm() = " << A.norm() * A.norm() << endl;
		cout << "Received_SINR = " << Received_SINR << endl;
		cout << "Received_SINR dBdB = " << 10.*log10(Received_SINR) << endl;
		getchar();
		*/


		if (Received_SINR)
			compute_CQI_offset(Received_SINR, rb_idx);
	}





	//cout << "Received_SINR = " << Received_SINR << endl;
	//getchar();



	return Received_SINR;

}

void MS::compute_CQI_offset(double Received_SINR, int rb_idx)
{
	num_added++;

	double difference = linear2dB(Received_SINR) - linear2dB(pppCQI_Map[receive_self_ms_idx][rb_idx][(cqi_history_length + t - SCHEDULE_DELAY) % cqi_history_length]);
	CQI_offset = (CQI_offset * (num_added - 1) + difference) / (double)num_added;
}


/*===================================================================
FUNCTION: UE::MESC_receiver_filter( )

DESCRIPTION:
===================================================================*/

MatrixXcd MS::MMSE_Receiver(int rb_idx, int stream_num)
{
	MatrixXcd H_bar, w, h;

	MatrixXcd Identity = MatrixXcd::Identity(NUM_RX_Port, NUM_RX_Port);


	double geometry = link[receive_self_ms_idx].str_signal - link[receive_self_ms_idx].interference; // ̰Ŵ dB // ̰ (link[_self_idx].UE_PATH_LOSS / link[_self_idx].UE_ICI) ̰Ŷ  (linear)

	double linear_signal = pow(10., link[receive_self_ms_idx].str_signal / 10.);
	double linear_interference = pow(10., link[receive_self_ms_idx].interference / 10.);



	
	H_bar = sqrt(linear_signal) * H_m[rb_idx] * sector[link[receive_self_ms_idx]._sector_in_control].W[rb_idx];
/// Rx2                              RxT                    Tx2

	h = (H_bar.col(stream_num));
//  Rx1

	w = h.adjoint() * (H_bar * H_bar.adjoint() + linear_interference * Identity).inverse();
///1xR   1xR            Rx2       2xR                                     RxR 



	return w;

}

/*
void MS::MESC_Receiver_Filter(int rbs_idx)
{
	cmat A, B, C, w, u;
	double sinr_estimate;

	w.append_col(code_word[PMI[rbs_idx][(cqi_history_length + t - SCHEDULE_DELAY) % cqi_history_length]].get_row(0));

	A = UE_PATH_LOSS * H_m[rbs_idx] * w * w.H() * H_m[rbs_idx].H(); // (P_tx/num_transmit_antenna) *

	// Check if the A is used or not !!!!!!!!!!!!!!!!!!!!!!

	B = UE_PATH_LOSS * (H_m[rbs_idx] * (eye_c(num_transmit_antenna) - w * w.H()) * H_m[rbs_idx].H()); //(P_tx/num_transmit_antenna) *
	C = UE_ICI * eye_c(2);

	u = sqrt(UE_PATH_LOSS) * inv(C + B) * H_m[rbs_idx] * w; // sqrt( P_tx/num_transmit_antenna ) *
	u = u.H();

	return u;
}



void MS::MMSE_Reciever_2(int rb_idx, int stream_num)
{
	cmat H_bar, w, h;

	H_bar = sqrt(UE_PATH_LOSS) * H_m[rb_idx] * sector[_sector_in_control].W[rb_idx];

	h.append_col(H_bar.get_col(stream_num));

	cmat Z;
	Z.clear();
	for (int stream = 0; stream < 4; stream++)
	{
		if (stream != stream_num && ppSchedulerRead[rb_idx][stream].ue_selected > -1)
		{
			cmat v;
			v.append_col(H_bar.get_col(stream));
			Z += v;
		}
	}

	Z = Z * Z.H() + UE_ICI * eye_c(2);

	cmat Ryy = Z + h * h.H();

	w = inv(Ryy) * h; // h.H() * inv( H_bar * H_bar.H() + UE_ICI * eye_c(2) ); // (dBm2linear( THERMAL_NOISE_DENSITY ) * UE_BANDWIDTH ) 

	return w.H();
}
*/


/*===================================================================
FUNCTION: UE::compute_tone_SINR( )

DESCRIPTION:
===================================================================*/
double MS::compute_tone_SINR_MRC(int rb_idx)
{

	if (CH_CAL == 1) //// use run-time calculate
	{
		double Received_SINR = 0.;

		double  sum_of_ch_coef = 0.;

		for (int r_idx = 0; r_idx < NUM_RX_Port; r_idx++) // Received    Channel
		{
			for (int t_idx = 0; t_idx < NUM_TX_Port; t_idx++) // Transmitted Channel
			{
				sum_of_ch_coef += pow(abs(H_m[rb_idx](r_idx, t_idx)), 2.);

			}
		}


		double geometry = link[receive_self_ms_idx].str_signal - link[receive_self_ms_idx].interference; // ̰Ŵ dB // ̰ (link[_self_idx].UE_PATH_LOSS / link[_self_idx].UE_ICI) ̰Ŷ  (linear)

		double linear_signal = dBm2linear(link[receive_self_ms_idx].str_signal);
		double linear_interference = dBm2linear(link[receive_self_ms_idx].interference);

		linear_interference = linear_interference + noise;






		Received_SINR = sum_of_ch_coef * (linear_signal / linear_interference);   // 


		//Received_SINR = 2. * (linear_signal / linear_interference);   // 



		//cout << sum_of_ch_coef << endl;

		return Received_SINR;
	}
	else
	{

		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		//////////////////////////////// For Load Pre-calculate Parameter ///////////////////////////////////////////////////////
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		int los_idx;
		int indoor_idx;
		int v_idx;

		int SELF_BS_idx = (int)(link[receive_self_ms_idx]._sector_in_control / 3);

		if (channel[SELF_BS_idx][receive_self_ms_idx].LOS == 1)
		{
			los_idx = 1;
		}
		else
		{
			los_idx = 0;
		}

		if (Indoor == 1) //////// InH  
		{
			indoor_idx = 1;
		}
		else
		{
			indoor_idx = 0;
		}

		double distance = Get_distance(ms[receive_self_ms_idx].loc, bs[SELF_BS_idx].loc);

		if (TYPE == 1 || TYPE == 2 || TYPE == 3 || TYPE == 4 || TYPE == 5)
		{
			double angle_v = atan2((bs_height - ms_height), distance) * 180. / pi;

			//// 
			if (angle_v > MAX_v_angle - (MAX_v_angle / anglev_mapping_value))
			{
				v_idx = 2;
			}
			else if (angle_v < (MAX_v_angle / anglev_mapping_value))
			{
				v_idx = 0;
			}
			else
			{
				v_idx = 1;
			}
		}
		else
		{
			//v_idx = (int)(anglev_mapping_value * randnum.u());
			v_idx = 0;
		}
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////





		double Received_SINR = 0.;

		double  sum_of_ch_coef = 0.;


		sum_of_ch_coef = sum_of_ch_coef_precalc[t % precalculate_time][los_idx][indoor_idx][v_idx][rb_idx];



		double geometry = link[receive_self_ms_idx].str_signal - link[receive_self_ms_idx].interference; // ̰Ŵ dB // ̰ (link[_self_idx].UE_PATH_LOSS / link[_self_idx].UE_ICI) ̰Ŷ  (linear)

		double linear_signal = dBm2linear(link[receive_self_ms_idx].str_signal);
		double linear_interference = dBm2linear(link[receive_self_ms_idx].interference);

		linear_interference = linear_interference + noise;

		/*
		cout << "sum_of_ch_coef = " << sum_of_ch_coef << endl;
		cout << "link[self_ms_idx].UE_PATH_LOSS = " << link[self_ms_idx].UE_PATH_LOSS << endl;
		cout << "link[self_ms_idx].UE_ICI = " << link[self_ms_idx].UE_ICI << endl;
		getchar();
		*/

		Received_SINR = sum_of_ch_coef * (linear_signal / linear_interference);   // 


		//cout << sum_of_ch_coef << endl;


		return Received_SINR;
	}






}


